
#include "pglock.h"

#define kPspSize                512

//------------------------------------------------------------------------------
//      Hack_DosFragment ()
//              Fragment up the Dos memory into the sizes specified so that
//              any GlobalPageLock() will not be able to move the memory into
//              the Dos area.  This is usefull when some code that you do not
//              control will use GlobalPageLock() and does not know about this
//              Windows bug.
//
//      All of the pointers will be saved along the way so this operation
//      can be undone.  A pointer to this memory list will be returned and
//      should be passed to Hack_DosUnfragment().  The dwSize passed in
//      should be GetSelectorLimit() for best results or if not available a
//      number like 1K should do fine.
//------------------------------------------------------------------------------

LPVOID
Hack_DosFragment (DWORD dwSize)
{
        LPDWORD         pList = NULL;
        LPDWORD         pNext;
        DWORD           dwDosAlloc;
        DWORD           dwPspDosAlloc;


        // Make sure that a least one PSP is available.
        // That way if a LoadLibrary is trying to load a DLL that
        // has fixed segments declared in the def file, they will get
        // loaded high.
        // Also if we do this then the min memory size should be at least
        // this big.
        //
        // Do not try to worry about small amounts of memory.
        // Most code will not be affected by small blocks.
        // Must never go below 4 because pointers are stored in
        // memory allocated for the purpose of freeing the blocks later.
        if (kPspSize > dwSize)
        {
                return(pList);
        }

        dwPspDosAlloc = GlobalDosAlloc(kPspSize);

        while ( (dwDosAlloc=GlobalDosAlloc(dwSize)) != 0 )
        {
                pNext = (LPDWORD)(dwDosAlloc << 16);
                *pNext = (DWORD)pList;
                pList = pNext;
        }

        if (dwPspDosAlloc)
        {
                GlobalDosFree( LOWORD(dwPspDosAlloc) );
        }

        return(pList);
}


//------------------------------------------------------------------------------
//      Hack_DosUnfragment ()
//              Do the inverse of Hack_DosFragment().
//------------------------------------------------------------------------------

void
Hack_DosUnfragment (LPVOID pList)
{
        LPDWORD         pNext;


        while (pList)
        {
                pNext = pList;
                pList = (LPDWORD)*((LPDWORD)pList);
                GlobalDosFree( SELECTOROF(pNext) );
        }
}


//------------------------------------------------------------------------------
//      Hack_GlobalPageLock ()
//              Call this in place of GlobalPageLock().
//
//      Windows will eat all of Dos memory if GlobalPageLock() is called.
//      After that happens no other code can run because all Windows
//      apps need a 512 byte psp located in the Dos memory.
//
//      Windows has two calls we are interested in.
//              GlobalPageLock()
//                      This call will move the memory as low as it can and lock it.
//              GlobalDosAlloc()
//                      This call will call GlobalPageLock() and if the memory is not
//                      in the lower 1Meg Dos memory then fail.
//
//      Our code works because it eats all the Dos memory before
//      doing the requested memory lock.  Windows will lock the memory
//      but be unable to move the memory into the Dos area.  Then we
//      free all of our Dos memory.
//
//      Warning: If GlobalPageLock() is used on previousely GlobalPageLock'd
//      memory, you will still fail because Windows will still move the
//      locked memory.  Also if a linear address is cached it may now be invalid.
//------------------------------------------------------------------------------

UINT
Hack_GlobalPageLock (HGLOBAL SelMem)
{
        UINT            Return;
        LPDWORD         pList = NULL;
        DWORD           dwSize;


        dwSize = GetSelectorLimit((UINT)SelMem);

        // Fragment up the DOS memory.
        pList = Hack_DosFragment(dwSize);

        // PageLock our mem which should stay above the 1Meg line because
        // of all the fragments.
        Return = GlobalPageLock(SelMem);

        // Free our Dos fragments.
        Hack_DosUnfragment(pList);

        return(Return);
}

